@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2015 Canonical
+ * Copyright (C) 2011-2016 Canonical
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -67,7 +67,7 @@
 #define IDLE_THRESHOLD		(98)	/* Less than this and we assume the device is not idle */
 
 #define MAX_POWER_DOMAINS	(16)	/* Maximum number of power domains allowed */
-#define MAX_THERMAL_ZONES	(16)
+#define MAX_THERMAL_ZONES	(16)	/* Maximum number of thermal zones allowed */
 
 /* Histogram specific constants */
 #define MAX_DIVISIONS		(10)
@@ -77,6 +77,11 @@
 #define GOT_PPID		(0x02)
 #define GOT_ALL			(GOT_TGID | GOT_PPID)
 
+#define I915_ENERGY_UJ		"/sys/kernel/debug/dri/0/i915_energy_uJ"
+
+#define MAX(x, y)		(x) > (y) ? (x) : (y)
+#define MIN(x, y)		(x) > (y) ? (y) : (x)
+
 /* Statistics gathered from /proc/stat and process activity */
 typedef enum {
 	CPU_USER = 0,
@@ -96,6 +101,7 @@ typedef enum {
 	PROC_EXEC,
 	PROC_EXIT,
 	POWER_TOTAL,
+	POWER_GPU,
 	POWER_DOMAIN_0,
 	THERMAL_ZONE_0 = POWER_DOMAIN_0 + MAX_POWER_DOMAINS,
 	MAX_VALUES = THERMAL_ZONE_0 + MAX_THERMAL_ZONES,
@@ -105,14 +111,14 @@ typedef enum {
  *  C-State information, 1 for each unique C-state
  */
 typedef struct cpu_state {
+	struct cpu_state *hash_next;	/* next in hash table */
+	struct cpu_state *list_next;	/* linked list of C-states */
+	struct cpu_info  *cpu_info_list; /* linked list of CPUs that have this state */
 	char 		*name;		/* C-State name, e.g. C1E-IVB */
 	char		*name_short;	/* short name, e.g. C1E */
 	uint64_t	latency;	/* latency */
 	uint64_t	usage_total;	/* total usage count */
 	double		resident;	/* percentage time resident in this state */
-	struct cpu_state *hash_next;	/* next in hash table */
-	struct cpu_state *list_next;	/* linked list of C-states */
-	struct cpu_info  *cpu_info_list; /* linked list of CPUs that have this state */
 } cpu_state_t;
 
 /*
@@ -120,6 +126,8 @@ typedef struct cpu_state {
  * N-CPUs x N-CPU states
  */
 typedef struct cpu_info {
+	struct cpu_info	*hash_next;	/* Next in hash table */
+	struct cpu_info *list_next;	/* Next in list of cpu_infos list */
 	uint32_t	cpu_id;		/* CPU ID */
 	char		*state;		/* C-State sysfs name, e.g 'state2' */
 	cpu_state_t 	*cpu_state;	/* C-State info */
@@ -132,8 +140,6 @@ typedef struct cpu_info {
 	uint64_t	prev_usage;	/* Previous usage count */
 	uint64_t	usage;		/* Current usage count */
 	uint64_t	usage_diff;	/* Difference in usage count */
-	struct cpu_info	*hash_next;	/* Next in hash table */
-	struct cpu_info *list_next;	/* Next in list of cpu_infos list */
 } cpu_info_t;
 
 
@@ -177,6 +183,7 @@ static const char *cpu_path = "/sys/devices/system/cpu";
 #define OPTS_CPU_FREQ		(0x1000)	/* Average CPU frequency */
 #define OPTS_NO_STATS_HEADINGS	(0x2000)	/* No stats headings */
 #define OPTS_THERMAL_ZONE	(0x4000)	/* Thermal zones */
+#define OPTS_GPU		(0x8000)	/* graphics */
 
 #define OPTS_USE_NETLINK	(OPTS_SHOW_PROC_ACTIVITY | \
 				 OPTS_REDO_NETLINK_BUSY |  \
@@ -228,20 +235,20 @@ typedef struct {
 
 /* RAPL domain info */
 typedef struct rapl_info {
+	struct rapl_info *next;		/* Next RAPL domain */
 	char *name;			/* RAPL name */
 	char *domain_name;		/* RAPL domain name */
 	double max_energy_uj;		/* Energy in micro Joules */
 	double last_energy_uj;		/* Last energy reading in micro Joules */
 	double t_last;			/* Time of last reading */
 	bool is_package;		/* Is it a package? */
-	struct rapl_info *next;		/* Next RAPL domain */
 } rapl_info_t;
 
 /* Thermal zone info */
 typedef struct tz_info {
+	struct tz_info *next;		/* Next TZ */
 	char *name;			/* Thermal Zone pathname */
 	char *type;			/* Thermal Zone type */
-	struct tz_info *next;		/* Next TZ */
 } tz_info_t;
 
 #if defined(POWERSTAT_X86)
@@ -393,7 +400,7 @@ static int file_get_uint64(const char *const file, uint64_t *val)
  *  cpu_freq_format()
  *	scale cpu freq into a human readable form
  */
-static const char *cpu_freq_format(double freq)
+static const char *cpu_freq_format(const double freq)
 {
 	static char buffer[40];
 	char *suffix = "EHz";
@@ -408,7 +415,7 @@ static const char *cpu_freq_format(double freq)
 		}
 	}
 
-	snprintf(buffer, sizeof(buffer), "%5.2f %-5s",
+	snprintf(buffer, sizeof(buffer), "%5.2f %-3s",
 		freq / scale, suffix);
 
 	return buffer;
@@ -932,10 +939,16 @@ static void stats_headings(void)
 	uint8_t i;
 
 	if (opts & OPTS_USE_NETLINK)
-		printf("  Time    User  Nice   Sys  Idle    IO  Run Ctxt/s  IRQ/s Fork Exec Exit  Watts ");
+		printf("  Time    User  Nice   Sys  Idle    IO  Run Ctxt/s  IRQ/s Fork Exec Exit  Watts");
 	else
-		printf("  Time    User  Nice   Sys  Idle    IO  Run Ctxt/s  IRQ/s  Watts ");
+		printf("  Time    User  Nice   Sys  Idle    IO  Run Ctxt/s  IRQ/s  Watts");
 
+	if (opts & (OPTS_DOMAIN_STATS |
+		    OPTS_THERMAL_ZONE |
+		    OPTS_CPU_FREQ |
+		    OPTS_GPU))
+		putchar(' ');
+		   
 	if (opts & OPTS_DOMAIN_STATS) {
 		for (i = 0; i < power_domains; i++)
 			printf(" %6.6s",
@@ -947,6 +960,8 @@ static void stats_headings(void)
 	}
 	if (opts & OPTS_CPU_FREQ)
 		printf(" %9.9s", "CPU Freq");
+	if (opts & OPTS_GPU)
+		printf(" %6.6s", "GPU W");
 	printf("\n");
 }
 
@@ -959,9 +974,9 @@ static void stats_ruler(void)
 	uint8_t i;
 
 	if (opts & OPTS_USE_NETLINK)
-		printf("-------- ----- ----- ----- ----- ----- ---- ------ ------ ---- ---- ---- ------");
+		printf("-------- ----- ----- ----- ----- ----- ---- ------ ------ ---- ---- ---- ------ ");
 	else
-		printf("-------- ----- ----- ----- ----- ----- ---- ------ ------ ------");
+		printf("-------- ----- ----- ----- ----- ----- ---- ------ ------ ------ ");
 
 	if (opts & OPTS_DOMAIN_STATS) {
 		for (i = 0; i < power_domains; i++)
@@ -973,6 +988,8 @@ static void stats_ruler(void)
 	}
 	if (opts & OPTS_CPU_FREQ)
 		printf(" ---------");
+	if (opts & OPTS_GPU)
+		printf(" ------");
 	printf("\n");
 }
 
@@ -1007,9 +1024,9 @@ static void stats_print(
 
 	if (summary) {
 		if (s->inaccurate[POWER_TOTAL])
-			(void)snprintf(buf, sizeof(buf), " -N/A-");
+			(void)snprintf(buf, sizeof(buf), " -N/A- ");
 		else
-			(void)snprintf(buf, sizeof(buf), "%6.2f",
+			(void)snprintf(buf, sizeof(buf), "%6.2f ",
 				s->value[POWER_TOTAL]);
 	} else {
 		(void)snprintf(buf, sizeof(buf), "%6.2f%s",
@@ -1055,6 +1072,12 @@ static void stats_print(
 	}
 	if (opts & OPTS_CPU_FREQ)
 		printf(" %s", cpu_freq_format(s->value[CPU_FREQ]));
+	if (opts & OPTS_GPU) {
+		if (s->inaccurate[POWER_GPU])
+			printf("  -N/A-");
+		else
+			printf(" %6.2f", s->value[POWER_GPU]);
+	}
 	printf("\n");
 }
 
@@ -1088,7 +1111,6 @@ static void stats_average_stddev_min_max(
 				valid++;
 			}
 		}
-
 		if (valid) {
 			average->value[j] = total / (double)valid;
 			total = 0.0;
@@ -1580,7 +1602,7 @@ static int power_get_proc_acpi(
  *  rapl_free_list()
  *	free RAPL list
  */
-void rapl_free_list(void)
+static void rapl_free_list(void)
 {
 	rapl_info_t *rapl = rapl_list;
 
@@ -1594,6 +1616,10 @@ void rapl_free_list(void)
 	}
 }
 
+/*
+ *  rapl_get_domain()
+ *	get RAPL domain name for nth domain
+ */
 static const char *rapl_get_domain(const int n)
 {
 	int i;
@@ -1685,7 +1711,7 @@ static int rapl_get_domains(void)
 		n++;
 	}
 	(void)closedir(dir);
-	power_domains = n;
+	power_domains = MIN(n, MAX_POWER_DOMAINS);
 
 	if (!n)
 		printf("Device does not have any RAPL domains, cannot power measure power usage.\n");
@@ -1699,10 +1725,10 @@ static int rapl_get_domains(void)
 static char *power_get_rapl_domain_names(void)
 {
 	char *names = NULL;
-	size_t len = 0;
+	size_t len = 0, n = 0;
 	rapl_info_t *rapl;
 
-	for (rapl = rapl_list; rapl; rapl = rapl->next) {
+	for (rapl = rapl_list; rapl && (n < MAX_POWER_DOMAINS); rapl = rapl->next, n++) {
 		char new_name[strlen(rapl->domain_name) + 3];
 		char *tmp;
 		size_t new_len;
@@ -1745,7 +1771,7 @@ static int power_get_rapl(
 
 	t_now = gettime_to_double();
 
-	for (rapl = rapl_list; rapl; rapl = rapl->next) {
+	for (rapl = rapl_list; rapl && (n < MAX_POWER_DOMAINS); rapl = rapl->next) {
 		char path[PATH_MAX];
 		FILE *fp;
 		double ujoules;
@@ -1773,12 +1799,13 @@ static int power_get_rapl(
 
 			if (first || (t_delta <= 0.0)) {
 				stats->value[POWER_DOMAIN_0 + n] = 0.0;
-				stats->inaccurate[POWER_TOTAL] = true;
+				stats->inaccurate[POWER_DOMAIN_0 + n] = true;
 			} else {
 				stats->value[POWER_DOMAIN_0 + n] =
 					(ujoules - last_energy_uj) / (t_delta * 1000000.0);
+				stats->inaccurate[POWER_DOMAIN_0 + n] = false;
 			}
-			if (rapl->is_package)
+			if (rapl->is_package && !stats->inaccurate[POWER_DOMAIN_0 + n])
 				stats->value[POWER_TOTAL] += stats->value[POWER_DOMAIN_0 + n];
 			n++;
 			*discharging = true;
@@ -1803,12 +1830,62 @@ static int power_get_rapl(
 }
 #endif
 
+/*
+ *  power_get_gpu_i915()
+ *	get Intel i915 GPU power stats
+ */
+static int power_get_gpu_i915(
+	const stats_t *const s1,
+	stats_t *s2,
+	stats_t *const res)
+{
+#if defined(POWERSTAT_X86)
+	uint64_t val;
+	int ret;
+
+	if (file_get_uint64(I915_ENERGY_UJ, &val) < 0) {
+		s2->value[POWER_GPU] = 0.0;
+		res->value[POWER_GPU] = 0.0;
+		res->inaccurate[POWER_GPU] = true;
+		ret = -1;
+	} else {
+		s2->value[POWER_GPU] = (double)val / 1000000.0;
+		res->value[POWER_GPU] = stats_sane(s1, s2, POWER_GPU);
+		res->inaccurate[POWER_GPU] = false;
+		ret = 0;
+	}
+	return ret;
+#else
+	(void)s1;
+	(void)s2;
+	(void)res;
+
+	return -1;
+#endif
+}
+
+/*
+ *  power_get_gpu()
+ *	get generic GPU power stats
+ */
+static void power_get_gpu(
+	const stats_t *const s1,
+	stats_t *s2,
+	stats_t *const res)
+{
+
+	if (power_get_gpu_i915(s1, s2, res) < 0) {
+		s2->value[POWER_GPU] = 0.0;
+		res->value[POWER_GPU] = 0.0;
+		res->inaccurate[POWER_GPU] = true;
+	}
+}
 
 /*
  *  tz_free_list()
  *	free thermal zone list
  */
-void tz_free_list(void)
+static void tz_free_list(void)
 {
 	tz_info_t *tz = tz_list;
 
@@ -1902,7 +1979,7 @@ static int tz_get_zones(void)
 		n++;
 	}
 	(void)closedir(dir);
-	thermal_zones = n;
+	thermal_zones = MIN(n, MAX_THERMAL_ZONES);
 
 	return n;
 }
@@ -1911,12 +1988,12 @@ static int tz_get_zones(void)
  *  tz_get_temperature()
  *	get temperatures from thermal zones
  */
-static int tz_get_temperature( stats_t *stats)
+static int tz_get_temperature(stats_t *const stats)
 {
 	tz_info_t *tz;
 	int n = 0;
 
-	for (tz = tz_list; tz; tz = tz->next) {
+	for (tz = tz_list; tz && (n < thermal_zones); tz = tz->next) {
 		char path[PATH_MAX];
 		uint64_t temp;
 
@@ -1947,8 +2024,8 @@ static int power_get(
 	int i;
 
 	for (i = POWER_TOTAL; i < POWER_DOMAIN_0 + MAX_POWER_DOMAINS; i++) {
-		stats->value[i] = i;
-		stats->inaccurate[i] = 0.0;
+		stats->value[i] = 0.0;
+		stats->inaccurate[i] = true;
 	}
 
 #if defined(POWERSTAT_X86)
@@ -1969,28 +2046,29 @@ static int power_get(
 }
 
 /*
- *  pjw()
- *	Hash a string, from Aho, Sethi, Ullman, Compiling Techniques.
+ *  hash_djb2a()
+ *	Hash a string, from Dan Bernstein comp.lang.c (xor version)
  */
-static uint32_t pjw(const char *str, const uint32_t id)
+static uint32_t hash_djb2a(const char *str, const uint32_t id)
 {
-	uint32_t h = id * 173;
+	register uint32_t hash = 5381 + id;
+	register int c;
 
-	while (*str) {
-		uint32_t g;
-		h = (h << 4) + (*str);
-		if (0 != (g = h & 0xf0000000)) {
-			h = h ^ (g >> 24);
-			h = h ^ g;
-		}
-		str++;
+	while ((c = *str++)) {
+		/* (hash * 33) ^ c */
+		hash = ((hash << 5) + hash) ^ c;
 	}
-	return h;
+	return hash;
 }
 
+/*
+ *  cpu_state_get()
+ *	get CPU C state based on name. We try to cache
+ * 	these in cpu_states to make lookup faster
+ */
 static cpu_state_t *cpu_state_get(const char *name)
 {
-	uint32_t h = pjw(name, 0) % MAX_STATES;
+	uint32_t h = hash_djb2a(name, 0) % MAX_STATES;
 	cpu_state_t *s = cpu_states[h];
 	size_t len;
 	char *ptr;
@@ -2025,9 +2103,13 @@ static cpu_state_t *cpu_state_get(const char *name)
 	return s;
 }
 
+/*
+ *  cpu_info_get()
+ *	get per CPU active state information
+ */
 static cpu_info_t *cpu_info_get(const char *state, const uint32_t cpu_id)
 {
-	uint32_t h = pjw(state, cpu_id) % MAX_CPUS;
+	uint32_t h = hash_djb2a(state, cpu_id) % MAX_CPUS;
 	cpu_info_t *ci = cpu_info[h];
 	FILE *fp;
 	char path[PATH_MAX];
@@ -2085,6 +2167,11 @@ static cpu_info_t *cpu_info_get(const char *state, const uint32_t cpu_id)
 	return ci;
 }
 
+
+/*
+ *  cpu_info_update()
+ *	update CPU state info
+ */
 static int cpu_info_update(cpu_info_t *ci)
 {
 	char path[PATH_MAX];
@@ -2120,6 +2207,10 @@ static int cpu_info_update(cpu_info_t *ci)
 	return 0;
 }
 
+/*
+ *  cpu_states_update()
+ *	update CPU state info for all CPUs
+ */
 static void cpu_states_update(void)
 {
 	struct dirent **cpu_list;
@@ -2164,6 +2255,10 @@ static void cpu_states_update(void)
 		free(cpu_list);
 }
 
+/*
+ *  cpu_states_free()
+ *	free cache of CPU state info
+ */
 static void cpu_states_free(void)
 {
 	cpu_state_t *s = cpu_states_list;
@@ -2187,6 +2282,10 @@ static void cpu_states_free(void)
 	}
 }
 
+/*
+ *  cpu_states_dump()
+ *	dump out gather CPU state statistics
+ */
 static void cpu_states_dump(void)
 {
 	cpu_state_t *s;
@@ -2439,6 +2538,8 @@ static int monitor(const int sock)
 	}
 	if (opts & OPTS_CSTATES)
 		cpu_states_update();
+	if (opts & OPTS_GPU)
+		power_get_gpu(&s2, &s1, &stats[readings]);
 
 	while (!stop_recv && (readings < max_readings)) {
 		double time_now, secs;
@@ -2514,6 +2615,7 @@ static int monitor(const int sock)
 				continue;
 			}
 
+
 			if ((opts & OPTS_REDO_WHEN_NOT_IDLE) &&
 			    (!stats[readings].inaccurate[CPU_IDLE]) &&
 			    (stats[readings].value[CPU_IDLE] < idle_threshold)) {
@@ -2533,6 +2635,9 @@ static int monitor(const int sock)
 			if (opts & OPTS_THERMAL_ZONE)
 				tz_get_temperature(&stats[readings]);
 
+			if (opts & OPTS_GPU)
+				power_get_gpu(&s1, &s2, &stats[readings]);
+
 			if (!discharging) {
 				free(stats);
 				return -1;	/* No longer discharging! */
@@ -2659,8 +2764,20 @@ static int monitor(const int sock)
 	}
 
 	printf("Summary:\n");
-	printf("%6.2f Watts on average with standard deviation %-6.2f\n",
-		average.value[POWER_TOTAL], stddev.value[POWER_TOTAL]);
+#if defined(POWERSTAT_X86)
+	if (opts & OPTS_RAPL) {
+		printf("CPU: %6.2f Watts on average with standard deviation %-6.2f\n",
+			average.value[POWER_TOTAL], stddev.value[POWER_TOTAL]);
+	} else
+#endif
+	{
+		printf("System: %6.2f Watts on average with standard deviation %-6.2f\n",
+			average.value[POWER_TOTAL], stddev.value[POWER_TOTAL]);
+	}
+	if (opts & OPTS_GPU) {
+		printf("GPU: %6.2f Watts on average with standard deviation %-6.2f\n",
+			average.value[POWER_GPU], stddev.value[POWER_GPU]);
+	}
 
 #if defined(POWERSTAT_X86)
 	if (opts & OPTS_RAPL) {
@@ -2726,7 +2843,7 @@ static int monitor(const int sock)
  *  show_help()
  *	simple help
  */
-void show_help(char *const argv[])
+static void show_help(char *const argv[])
 {
 	printf("%s, version %s\n\n", app_name, VERSION);
 	printf("usage: %s [options] [delay [count]]\n", argv[0]);
@@ -2736,6 +2853,7 @@ void show_help(char *const argv[])
 	printf("\t-d specify delay before starting, default is %" PRId32 " seconds\n", start_delay);
 	printf("\t-D show RAPL domain power measurements (enables -R option)\n");
 	printf("\t-f show average CPU frequency\n");
+	printf("\t-g show GPU power (currently just i915)\n");
 	printf("\t-h show help\n");
 	printf("\t-H show spread of measurements with power histogram\n");
 	printf("\t-i specify CPU idle threshold, used in conjunction with -b\n");
@@ -2761,11 +2879,14 @@ int main(int argc, char * const argv[])
 	struct sigaction new_action;
 	stats_t dummy_stats;
 
+	if (geteuid() == 0)
+		opts |= OPTS_ROOT_PRIV;
+
 	for (;;) {
 #if defined(POWERSTAT_X86)
-		int c = getopt(argc, argv, "abd:cDfhHi:nprszStR");
+		int c = getopt(argc, argv, "abd:cDfghHi:nprszStR");
 #else
-		int c = getopt(argc, argv, "abd:cDfhHi:nprszSt");
+		int c = getopt(argc, argv, "abd:cDfghHi:nprszSt");
 #endif
 		if (c == -1)
 			break;
@@ -2775,6 +2896,8 @@ int main(int argc, char * const argv[])
 				 OPTS_CPU_FREQ |
 				 OPTS_HISTOGRAM |
 				 OPTS_THERMAL_ZONE);
+			if (opts & OPTS_ROOT_PRIV)
+				 opts |= OPTS_GPU;
 			break;
 		case 'b':
 			opts |= OPTS_REDO_WHEN_NOT_IDLE;
@@ -2801,6 +2924,9 @@ int main(int argc, char * const argv[])
 		case 'f':
 			opts |= OPTS_CPU_FREQ;
 			break;
+		case 'g':
+			opts |= OPTS_GPU;
+			break;
 		case 'h':
 			show_help(argv);
 			exit(EXIT_SUCCESS);
@@ -2907,10 +3033,9 @@ int main(int argc, char * const argv[])
 	if (max_readings < 10)
 		fprintf(stderr, "Number of readings low, results may be inaccurate.\n");
 
-	if (geteuid() == 0)
-		opts |= OPTS_ROOT_PRIV;
-	else if (opts & OPTS_USE_NETLINK) {
-		fprintf(stderr, "%s needs to be run with root privilege when using -p, -r, -s options.\n", argv[0]);
+	if (!(opts & OPTS_ROOT_PRIV) &&
+	    (opts & (OPTS_USE_NETLINK | OPTS_GPU))) {
+		fprintf(stderr, "%s needs to be run with root privilege when using -g, -p, -r, -s options.\n", argv[0]);
 		exit(ret);
 	}
 
